<?php
namespace com\dunweber\contactform;
// ========================================================================
// PHP ContactForm
// ------------------------------------------------------------------------
// https://www.dunweber.com/docs/scripts/releases/contactform
// Copyright (C) 2021 Christian L. Dünweber
// MIT License: https://opensource.org/licenses/MIT
// ------------------------------------------------------------------------
// See README for instructions.
// ========================================================================

// ========================================================================
// Configuration
// ------------------------------------------------------------------------
class Conf
{
	// reCAPTCHA v3 site key used in the HTML code
	public const RECAP_SITE_KEY = 'xxx';

	// reCAPTCHA v3 secret key for communication between server and reCAPTCHA
	public const RECAP_SECRET_KEY = 'yyy';

	// reCAPTCHA v3 score threshold: 0.0 (bad) to 1.0 (good)
	public const RECAP_THRESHOLD = 0.5;

	// Language: Filename without extension placed in subfolder lang, e.g. en_US
	public const LANGUAGE = 'en_US';

	// Date and time format for timestamps
	public const DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S (%z)';

	// Set desired timezone used for time
	public const DATETIME_TIMEZONE = 'Europe/Copenhagen';

	// Character encoding used for the webpage, usually UTF-8
	public const ENCODING = 'UTF-8';

	// Web path to stylesheet to include on the page
	public const FORMCSSPATH = 'style/style.form.css';

	// Web path to stylesheet to include in the mail
	public const MAILCSSPATH = 'style/style.mail.css';
}

// Include language file
if (!file_exists('lang/'.Conf::LANGUAGE.'.php')) {
	exit('ContactForm: Language file not found: '.Conf::LANGUAGE.'.php');
}

require_once('lang/'.Conf::LANGUAGE.'.php');

// ========================================================================
// ContactForm Implementation.
// ------------------------------------------------------------------------
class ContactForm
{
	private $opts = [];

	// --------------------------------------------------------------------
	// Constructor. Provide array of options for the form.
	// --------------------------------------------------------------------
	public function __construct(array $opts)
	{
		$this->opts = $opts;
		ini_set('default_charset', Conf::ENCODING);
		mb_internal_encoding(Conf::ENCODING);
		date_default_timezone_set(Conf::DATETIME_TIMEZONE);

		if (!isset($this->opts['receivers'], $this->opts['sender'])) {
			exit('ContactForm: Specify sender and receiver(s) e-mail addresses via $opts');
		}
	}

	// --------------------------------------------------------------------
	// Get the complete contact form and success or error message.
	// @return HTML string.
	// --------------------------------------------------------------------
	public function getContactForm(): string
	{
		// Check if the form has been properly submitted
		if (!isset($_POST['reCapToken'])) {
			return $this->getSystemWrapper($this->getForm());
		}

		// Get submitted data
		$name = $this->post2html($_POST['name'] ?? '', true);
		$email = $this->post2html($_POST['email'] ?? '', true);
		$subject = $this->post2html($_POST['subject'] ?? '', true);
		$message = nl2br($this->post2html($_POST['message'] ?? '', false));
		$copy = isset($_POST['copy']) ? true : false;
		$out = '';

		// Check captcha
		if (!$this->isCaptchaPassed()) {
			$out = $this->showMessage(Lang::ERROR_TITLE, Lang::ERROR_GENERIC, 'errorText');
		}
		else {
			// Validate the data
			$errors = $this->validateData($name, $email, $subject, $message);

			// Send the mail and print a status message or explain errors to the user
			if (count($errors) > 0) {
				$out = $this->showMessage(Lang::ERROR_TITLE, implode('<br/>', $errors), 'errorText');
			}
			else {
				$out = $this->sendMail($name, $email, $subject, $message, $copy, $this->opts['receivers'], $this->opts['sender']);
			}
		}

		return $this->getSystemWrapper($this->getForm() . $out);
	}

	// --------------------------------------------------------------------
	// Complete HTML contact form possibly with user submitted data posted
	// back to save for prompting to correct in case of errors.
	// @return HTML string.
	// --------------------------------------------------------------------
	private function getForm(): string
	{
		return '
		<form id="cfForm" action="#cfAnchor" method="post" onsubmit="cf.submitForm(event)" accept-charset="'.Conf::ENCODING.'">
		<fieldset>
		<legend>'.Lang::FORM_TITLE.'</legend>
			<div class="row">
			<label for="name">'.Lang::FORM_SENDER_NAME.'</label>
			<input type="text" id="name" name="name" onkeypress="return event.keyCode!=13" required="required"
				value="'.$this->post2html($_POST['name'] ?? '').'" />
			</div>

			<div class="row">
			<label for="email">'.Lang::FORM_SENDER_EMAIL.'</label>
			<input type="email" id="email" name="email" onkeypress="return event.keyCode!=13" required="required"
				value="'.$this->post2html($_POST['email'] ?? '').'" />
			</div>

			<div class="row">
			<label for="subject">'.Lang::FORM_SENDER_SUBJECT.'</label>
			<input type="text" id="subject" name="subject" onkeypress="return event.keyCode!=13" required="required"
				value="'.$this->post2html($_POST['subject'] ?? '').'" />
			</div>

			<div class="row">
			<label for="message">'.Lang::FORM_SENDER_MESSAGE.'</label>
			<textarea id="message" name="message" required="required">'.$this->post2html($_POST['message'] ?? '').'</textarea>		
			</div>

			<div class="row copyRow">
			<label for="copy" class="copyLabel">'.Lang::FORM_COPY.'</label>
			<input type="checkbox" id="copy" name="copy" value="1"
				onkeypress="return event.keyCode!=13" '.(isset($_POST["copy"]) ? 'checked="checked "' : '').'/>
			<button id="submitButton" class="submitButton" value="'.Lang::FORM_BUTTON_SEND.'">'.Lang::FORM_BUTTON_SEND.'</button>
			</div>

			<input type="hidden" id="reCapSiteKey" value="'.Conf::RECAP_SITE_KEY.'" />
			<input type="hidden" id="reCapAction" value="'.($this->opts['reCapAction'] ?? 'ContactFormSubmit').'" />
			<input type="hidden" id="reCapToken" name="reCapToken" />
			<input type="hidden" id="errorTitle" value="'.Lang::ERROR_TITLE.'" />
			<input type="hidden" id="errorConnection" value="'.Lang::ERROR_CONNECTION.'" />
			<input type="hidden" id="errorConfig" value="'.Lang::ERROR_CONFIG.'" />
			<input type="hidden" id="errorGeneric" value="'.Lang::ERROR_GENERIC.'" />
		</fieldset>
		</form>

		<p><em><small>This form is protected by reCAPTCHA and the Google <a href="https://policies.google.com/privacy">Privacy Policy</a>
		and <a href="https://policies.google.com/terms">Terms of Service</a> apply.</small></em></p>';
	}

	// --------------------------------------------------------------------
	// reCAPTCHA: Check that user seems to not be a robot.
	// --------------------------------------------------------------------
	private function isCaptchaPassed(): bool
	{
		$postData = http_build_query(
			['secret' => Conf::RECAP_SECRET_KEY,
			 'response' => $_POST['reCapToken'] ?? '',
			 'remoteip' => $_SERVER['REMOTE_ADDR']]
		);

		$options = ['http' =>
			['method'  => 'POST',
			 'header'  => 'Content-Type: application/x-www-form-urlencoded',
			 'content' => $postData]];

		$req = file_get_contents('https://www.google.com/recaptcha/api/siteverify', false, stream_context_create($options));
		$res = @json_decode($req);
		return $res->success && ($res->score >= Conf::RECAP_THRESHOLD);
	}

	// --------------------------------------------------------------------
	// Validate the submitted data.
	// @params Data to validate.
	// @return Array of error messages or empty array if validated.
	// --------------------------------------------------------------------
	public function validateData(string $name, string $email, string $subject, string $message): array
	{
		$errors = [];

		if (strlen($name) == 0) {
			$errors[] = Lang::ERROR_MISSING_NAME;
		}
		if (!$this->checkEmail($email)) {
			$errors[] = Lang::ERROR_INVALID_EMAIL;
		}
		if (strlen($subject) == 0) {
			$errors[] = Lang::ERROR_MISSING_SUBJECT;
		}
		if (strlen($message) == 0) {
			$errors[] = Lang::ERROR_MISSING_MESSAGE;
		}

		return $errors;
	}

	// --------------------------------------------------------------------
	// Send mail as HTML to a list of receivers.
	// @param $name Name of the sender.
	// @param $email E-mail address of the sender.
	// @param $subject Subject of the mail.
	// @param $message Body of the mail.
	// @param $copy True or false for sending a self-copy to $email.
	// @param $rxEmails Comma separated list of receiver e-mail addresses.
	// @param $txEmail Sender e-mail addresses.
	// @return HTML string with success or error message.
	// --------------------------------------------------------------------
	private function sendMail(string $name, string $email, string $subject, string $message, bool $copy, string $rxEmails, string $txEmail): string
	{
		// Build HTML document for sending
		$bodyAdm = $this->getHTMLmail($name, $email, $subject, $message, '<p>'.($copy ? 'User selected a copy' : 'User did not select a copy').'.</p>');
		$bodyCpy = $this->getHTMLmail($name, $email, $subject, $message, '');

		// Prepare to send e-mail: Headers must satisfy RFC 2047, message RFC 2822
		// Line length max 70 characters
		$bodyAdm = wordwrap($bodyAdm, 70);
		$bodyCpy = wordwrap($bodyCpy, 70);

		$subject = mb_encode_mimeheader(html_entity_decode($subject, ENT_QUOTES), Conf::ENCODING);
		$headers =
			'From: '.$txEmail. "\r\n".
			'MIME-Version: 1.0'. "\r\n".
			'Content-type: text/html; charset='.Conf::ENCODING. "\r\n".
			'Reply-To: '.$email. "\r\n".
			'X-Mailer: PHP/'.phpversion();
		
		// Send to all comma separated e-mails
		if (mail($rxEmails, $subject, $bodyAdm, $headers)) {
			if ($copy) {
				mail($email, $subject, $bodyCpy, $headers);
			}
			
			unset($_POST['name'], $_POST['email'], $_POST['subject'], $_POST['message'], $_POST['copy']);
			return $this->showMessage(Lang::SUCCESS_TITLE, Lang::SUCCESS_MESSAGE, 'successText');
		}
		else {
			return $this->showMessage(Lang::ERROR_TITLE, Lang::ERROR_GENERIC, 'errorText');
		}
	}

	// --------------------------------------------------------------------
	// Build HTML document for sending as mail.
	// @params Data to tabulize.
	// @return HTML document as string.
	// --------------------------------------------------------------------
	private function getHTMLmail(string $name, string $email, string $subject, string $message, string $copyInfo): string
	{
		return
			'<!DOCTYPE html>
			<html lang="en">
			<head>
				<title>'.Lang::FORM_TITLE.' '.$_SERVER['HTTP_HOST'].'</title>
				<meta charset="'.Conf::ENCODING.'" />
				<style>'.(file_get_contents(Conf::MAILCSSPATH) ?? '').'</style>
			</head>
			<body>
				<h1>'.Lang::FORM_TITLE.' '.$_SERVER['HTTP_HOST'].'</h1>
				<table>
					<tr>
						<td class="leftCol">'.Lang::FORM_SENDER_NAME.'</td>
						<td>'.$name.'</td>
					</tr>
					<tr>
						<td class="leftCol">'.Lang::FORM_SENDER_EMAIL.'</td>
						<td>'.$email.'</td>
					</tr>
					<tr>
						<td class="leftCol">'.Lang::FORM_SENDER_SUBJECT.'</td>
						<td>'.$subject.'</td>
					</tr>
					<tr>
						<td class="leftCol">'.Lang::FORM_SENDER_MESSAGE.'</td>
						<td>'.$message.'</td>
					</tr>
					<tr>
						<td class="leftCol">'.Lang::MAIL_SEND_TIME.'</td>
						<td>'.strftime(Conf::DATETIME_FORMAT).'</td>
					</tr>
					<tr>
						<td class="leftCol">'.Lang::MAIL_SENDER_IP.'</td>
						<td>'.$_SERVER['REMOTE_ADDR'].'</td>
					</tr>
				</table>
				'.$copyInfo.'
			</body>
			</html>';
	}

	// --------------------------------------------------------------------
	// Show message to the user via JS wrapper.
	// @param Strings with title, message, and CSS class name.
	// @return HTML string calling JS.
	// --------------------------------------------------------------------
	private function showMessage(string $title, string $message, string $css): string
	{
		return '<script>cf.showMessage("'.$title.'", "'.$message.'", "'.$css.'");</script>';
	}

	// --------------------------------------------------------------------
	// Get the system display wrapper.
	// @param $content String HTML content to wrap.
	// @return String with input wrapped in the system outer HTML container.
	// --------------------------------------------------------------------
	public function getSystemWrapper(string $content): string
	{
		return '<div class="cfContainer"><span id="cfAnchor">&nbsp;</span>
		<script src="https://www.google.com/recaptcha/api.js?render='.Conf::RECAP_SITE_KEY.'" async defer></script>
		<script src="ContactForm.js"></script>
		<script>const cf = new ContactForm();</script>
		'.$content.'</div>';
	}

	// --------------------------------------------------------------------
	// Trim and pass through htmlspecialchars using specified encoding.
	// Use for data from posted form for posting back in the form or
	// anything else to be shown that may contain HTML characters.
	// @param $str String to process.
	// @return String converted and trimmed.
	// --------------------------------------------------------------------
	private function post2html(string $str, bool $stripNewlines = false): string
	{
		$str = htmlspecialchars($str, ENT_QUOTES|ENT_SUBSTITUTE|ENT_HTML5, Conf::ENCODING);
		return trim($stripNewlines ? preg_replace("/\r|\n/", ' ', $str) : $str);
	}

	// --------------------------------------------------------------------
	// Check an e-mail address to be on the form x@x.xx or x@x.xxx and
	// check for a DNS record of type MX, A, or CNAME for given Internet
	// host name (only works on Windows as of PHP 5.3.0).
	// @param $email String to check.
	// @return boolean
	// --------------------------------------------------------------------
	private function checkEmail(string $email): bool
	{ 
		if (preg_match('/^.+\@(\[?)[a-zA-Z0-9\-\.]+\.([a-zA-Z]{2,3}|[0-9]{1,3})(\]?)$/u',$email) === 1) {
			$host = explode('@', $email);
			
			if (function_exists('checkdnsrr')) {
				if (checkdnsrr($host[1], 'MX') ||
					checkdnsrr($host[1], 'A') ||
					checkdnsrr($host[1], 'CNAME')) {
					return true;
				}
			}
			else {
				return true; // Do not block prior to PHP 5.3.0 on Windows
			}
		}
		
		return false;
	}
}

?>
